/*  ---------------------------------------------------------------  */
/*           The HMM-Based Speech Synthesis System (HTS)             */
/*                       HTS Working Group                           */
/*                                                                   */
/*                  Department of Computer Science                   */
/*                  Nagoya Institute of Technology                   */
/*                               and                                 */
/*   Interdisciplinary Graduate School of Science and Engineering    */
/*                  Tokyo Institute of Technology                    */
/*                                                                   */
/*                     Copyright (c) 2001-2007                       */
/*                       All Rights Reserved.                        */
/*                                                                   */
/*  Permission is hereby granted, free of charge, to use and         */
/*  distribute this software and its documentation without           */
/*  restriction, including without limitation the rights to use,     */
/*  copy, modify, merge, publish, distribute, sublicense, and/or     */
/*  sell copies of this work, and to permit persons to whom this     */
/*  work is furnished to do so, subject to the following conditions: */
/*                                                                   */
/*    1. The source code must retain the above copyright notice,     */
/*       this list of conditions and the following disclaimer.       */
/*                                                                   */
/*    2. Any modifications to the source code must be clearly        */
/*       marked as such.                                             */
/*                                                                   */
/*    3. Redistributions in binary form must reproduce the above     */
/*       copyright notice, this list of conditions and the           */
/*       following disclaimer in the documentation and/or other      */
/*       materials provided with the distribution.  Otherwise, one   */
/*       must contact the HTS working group.                         */
/*                                                                   */
/*  NAGOYA INSTITUTE OF TECHNOLOGY, TOKYO INSTITUTE OF TECHNOLOGY,   */
/*  HTS WORKING GROUP, AND THE CONTRIBUTORS TO THIS WORK DISCLAIM    */
/*  ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL       */
/*  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT   */
/*  SHALL NAGOYA INSTITUTE OF TECHNOLOGY, TOKYO INSTITUTE OF         */
/*  TECHNOLOGY, HTS WORKING GROUP, NOR THE CONTRIBUTORS BE LIABLE    */
/*  FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY        */
/*  DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,  */
/*  WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTUOUS   */
/*  ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR          */
/*  PERFORMANCE OF THIS SOFTWARE.                                    */
/*                                                                   */
/*  ---------------------------------------------------------------  */

/* $Id: hts_engine.c,v 1.2 2007/09/28 05:39:41 zen Exp $ */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

#include "HTS_engine.h"

/* Usage: output usage */
void Usage(void)
{
   fprintf(stderr, "\n");
   fprintf(stderr, "hts_engine - A HMM-based speech synthesis engine\n");
   fprintf(stderr, "\n");
   fprintf(stderr, "  usage:\n");
   fprintf(stderr, "       hts_engine [ options ] [ infile ] \n");
   fprintf(stderr,
           "  options:                                                              [def] [min--max]\n");
   fprintf(stderr,
           "       -td tree      : decision trees file for state duration           [N/A]\n");
   fprintf(stderr,
           "       -tf tree      : decision trees file for Log F0                   [N/A]\n");
   fprintf(stderr,
           "       -tm tree      : decision trees file for Mel-Cepstrum             [N/A]\n");
   fprintf(stderr,
           "       -md pdf       : model file for state duration                    [N/A]\n");
   fprintf(stderr,
           "       -mf pdf       : model file for Log F0                            [N/A]\n");
   fprintf(stderr,
           "       -mm pdf       : model file for Mel-Cepstrum                      [N/A]\n");
   fprintf(stderr,
           "       -df win       : window file for culcuration delta of log F0      [N/A]\n");
   fprintf(stderr,
           "       -dm win       : filename of delta coeffs for Mel-Cepstrum        [N/A]\n");
   fprintf(stderr,
           "       -od s         : filename of output duration                      [N/A]\n");
   fprintf(stderr,
           "       -of s         : filename of output F0                            [N/A]\n");
   fprintf(stderr,
           "       -om s         : filename of output mcep                          [N/A]\n");
   fprintf(stderr,
           "       -or s         : filename of output raw audio (generated speech)  [N/A]\n");
   fprintf(stderr,
           "       -ow s         : filename of output wav audio (generated speech)  [N/A]\n");
   fprintf(stderr,
           "       -ot s         : filename of output trace information             [N/A]\n");
   fprintf(stderr,
           "       -vs           : use state alignment for duration                 [0]\n");
   fprintf(stderr,
           "       -vp           : use phoneme alignment for duration               [0]\n");
   fprintf(stderr,
           "       -i i f1 .. fi : enable interpolation & specify number(i),rate(f) [1][ 1-- ]\n");
   fprintf(stderr,
           "       -s i          : sampling frequency                               [16000][  0--48000]\n");
   fprintf(stderr,
           "       -p i          : frame period (point)                             [ 80 ][   0--2000]\n");
   fprintf(stderr,
           "       -a  f         : all-pass constant                                [0.42][ 0.0--1.0]\n");
   fprintf(stderr,
           "       -b  f         : postfiltering coefficient                        [0.00][-0.8--0.8]\n");
   fprintf(stderr,
           "       -r  f         : control duration parameter                       [0.00][-1.0--0.0]\n");
   fprintf(stderr,
           "       -fs f         : multilply f0                                     [1.00][ 0.0--5.0]\n");
   fprintf(stderr,
           "       -fm f         : add f0                                           [0.00][ 0.0--100.0]\n");
   fprintf(stderr,
           "       -u  f         : voiced/unvoiced threshold                        [0.50][ 0.0--1.0]\n");
   fprintf(stderr,
           "       -l  f         : length of generated speech (in second)           [N/A] [ 0.0--30.0]\n");
   fprintf(stderr, "  infile:\n");
   fprintf(stderr, "       label file\n");
   fprintf(stderr, "  note:\n");
   fprintf(stderr, "       option '-d' may be repeated to use multiple\n");
   fprintf(stderr, "       delta parameters\n");
   fprintf(stderr, "       if 'd' is an integer, delta coefficients are\n");
   fprintf(stderr, "       calculated based on a regression formula.\n\n");
   fprintf(stderr, "       generated mel-cepstrum and log F0 sequences are \n");
   fprintf(stderr, "       saved in natural endian, binary (float) format.\n");
   fprintf(stderr, "\n");

   exit(0);
}

/* Error: output error message */
void Error(const int error, char *message, ...)
{
   va_list arg;

   fflush(stdout);
   fflush(stderr);

   if (error > 0)
      fprintf(stderr, "\nError: ");
   else
      fprintf(stderr, "\nWarning: ");

   va_start(arg, message);
   vfprintf(stderr, message, arg);
   va_end(arg);

   fflush(stderr);

   if (error > 0)
      exit(error);
}

/* Getfp: wrapper for fopen */
FILE *Getfp(const char *name, const char *opt)
{
   FILE *fp = fopen(name, opt);

   if (fp == NULL)
      Error(2, "Getfp: Cannot open %s.\n", name);

   return (fp);
}

/* GetNumInterp: get number of speakers for interpolation from argv */
int GetNumInterp(int argc, char **argv_search)
{
   int num_interp = 1;
   while (--argc) {
      if (**++argv_search == '-') {
         if (*(*argv_search + 1) == 'i') {
            num_interp = atoi(*++argv_search);
            if (num_interp < 1) {
               num_interp = 1;
            }
            --argc;
         }
      }
   }
   return (num_interp);
}

int main(int argc, char **argv)
{
   int i;
   double f;
   FILE *labfp = stdin;
   FILE *lf0fp = NULL, *mcpfp = NULL, *durfp = NULL,
       *wavfp = NULL, *rawfp = NULL, *tracefp = NULL;

   /* number of speakers for interpolation */
   int num_interp = 0;
   double *rate_interp = NULL;

   /* file names of models */
   char **fn_ms_lf0;
   char **fn_ms_mcp;
   char **fn_ms_dur;
   /* number of each models for interpolation */
   int num_ms_lf0 = 0, num_ms_mcp = 0, num_ms_dur = 0;

   /* file names of trees */
   char **fn_ts_lf0;
   char **fn_ts_mcp;
   char **fn_ts_dur;
   /* number of each trees for interpolation */
   int num_ts_lf0 = 0, num_ts_mcp = 0, num_ts_dur = 0;

   /* file names of windows */
   char **fn_ws_lf0;
   char **fn_ws_mcp;
   int num_ws_lf0 = 0, num_ws_mcp = 0;

   HTS_Engine engine;
   HTS_Model m;

   /* parse command line */
   if (argc == 1) {
      Usage();
   }

   /* initialize */
   HTS_Engine_initialize(&engine);

   /* delta window handler for log f0 */
   fn_ws_lf0 = (char **) calloc(argc, sizeof(char *));
   /* delta window handler for mel-cepstrum */
   fn_ws_mcp = (char **) calloc(argc, sizeof(char *));

   /* prepare for interpolation */
   num_interp = GetNumInterp(argc, argv);
   rate_interp = (double *) calloc(num_interp, sizeof(double));
   for (i = 0; i < num_interp; i++)
      rate_interp[i] = 1.0;

   fn_ms_lf0 = (char **) calloc(num_interp, sizeof(char *));
   fn_ms_mcp = (char **) calloc(num_interp, sizeof(char *));
   fn_ms_dur = (char **) calloc(num_interp, sizeof(char *));
   fn_ts_lf0 = (char **) calloc(num_interp, sizeof(char *));
   fn_ts_mcp = (char **) calloc(num_interp, sizeof(char *));
   fn_ts_dur = (char **) calloc(num_interp, sizeof(char *));

   /* read command */
   while (--argc) {
      if (**++argv == '-') {
         switch (*(*argv + 1)) {
         case 'v':
            switch (*(*argv + 2)) {
            case 's':
               HTS_Engine_set_algnst(&engine, TRUE);
               break;
            case 'p':
               HTS_Engine_set_algnph(&engine, TRUE);
               break;
            default:
               Error(1, "hts_engine: Invalid option '-v%c'.\n", *(*argv + 2));
            }
            break;
         case 't':
            switch (*(*argv + 2)) {
            case 'f':
            case 'p':
               fn_ts_lf0[num_ts_lf0++] = *(argv + 1);
               break;
            case 'm':
               fn_ts_mcp[num_ts_mcp++] = *(argv + 1);
               break;
            case 'd':
               fn_ts_dur[num_ts_dur++] = *(argv + 1);
               break;
            default:
               Error(1, "hts_engine: Invalid option '-t%c'.\n", *(*argv + 2));
            }
            ++argv;
            --argc;
            break;
         case 'm':
            switch (*(*argv + 2)) {
            case 'f':
            case 'p':
               fn_ms_lf0[num_ms_lf0++] = *(argv + 1);
               break;
            case 'm':
               fn_ms_mcp[num_ms_mcp++] = *(argv + 1);
               break;
            case 'd':
               fn_ms_dur[num_ms_dur++] = *(argv + 1);
               break;
            default:
               Error(1, "hts_engine: Invalid option '-m%c'.\n", *(*argv + 2));
            }
            ++argv;
            --argc;
            break;
         case 'd':
            switch (*(*argv + 2)) {
            case 'f':
            case 'p':
               fn_ws_lf0[num_ws_lf0] = *(argv + 1);
               num_ws_lf0++;
               break;
            case 'm':
               fn_ws_mcp[num_ws_mcp] = *(argv + 1);
               num_ws_mcp++;
               break;
            default:
               Error(1, "hts_engine: Invalid option '-d%c'.\n", *(*argv + 2));
            }
            ++argv;
            --argc;
            break;
         case 'o':
            switch (*(*argv + 2)) {
            case 'w':
               wavfp = Getfp(*(argv + 1), "wb");
               break;
            case 'r':
               rawfp = Getfp(*(argv + 1), "wb");
               break;
            case 'f':
            case 'p':
               lf0fp = Getfp(*(argv + 1), "wb");
               break;
            case 'm':
               mcpfp = Getfp(*(argv + 1), "wb");
               break;
            case 'd':
               durfp = Getfp(*(argv + 1), "wt");
               break;
            case 't':
               tracefp = Getfp(*(argv + 1), "wt");
               break;
            default:
               Error(1, "hts_engine: Invalid option '-o%c'.\n", *(*argv + 2));
            }
            ++argv;
            --argc;
            break;
         case 'h':
            Usage();
            break;
         case 's':
            i = atoi(*++argv);
            HTS_Engine_set_sampling_rate(&engine, i);
            --argc;
            break;
         case 'p':
            i = atoi(*++argv);
            HTS_Engine_set_fperiod(&engine, i);
            --argc;
            break;
         case 'a':
            f = atof(*++argv);
            HTS_Engine_set_alpha(&engine, f);
            --argc;
            break;
         case 'b':
            f = atof(*++argv);
            HTS_Engine_set_beta(&engine, f);
            --argc;
            break;
         case 'r':
            f = atof(*++argv);
            HTS_Engine_set_rho(&engine, f);
            --argc;
            break;
         case 'f':
            switch (*(*argv + 2)) {
            case 's':
               f = atof(*++argv);
               HTS_Engine_set_f0_std(&engine, f);
               break;
            case 'm':
               f = atof(*++argv);
               HTS_Engine_set_f0_mean(&engine, f);
               break;
            default:
               Error(1, "hts_engine: Invalid option '-f%c'.\n", *(*argv + 2));
            }
            --argc;
            break;
         case 'u':
            f = atof(*++argv);
            HTS_Engine_set_uv(&engine, f);
            --argc;
            break;
         case 'l':
            f = atof(*++argv);
            HTS_Engine_set_length(&engine, f);
            --argc;
            break;
         case 'i':
            ++argv;
            argc--;
            for (i = 0; i < num_interp; i++) {
               rate_interp[i] = atof(*++argv);
               argc--;
            }
            break;
         default:
            Error(1, "hts_engine: Invalid option '-%c'.\n", *(*argv + 1));
         }
      } else {
         labfp = Getfp(*argv, "r");
      }
   }
   /* number of models,trees check */
   if (num_interp != num_ts_lf0 || num_interp != num_ts_mcp ||
       num_interp != num_ts_dur || num_interp != num_ms_lf0 ||
       num_interp != num_ms_mcp || num_interp != num_ms_dur) {
      Error(1,
            "hts_engine: specify %d models(trees) for each parameter.\n",
            num_interp);
   }

   /* load model, tree and windows */
   HTS_Engine_load_fn(&engine,
                      fn_ms_lf0, fn_ms_mcp, fn_ms_dur,
                      fn_ts_lf0, fn_ts_mcp, fn_ts_dur,
                      num_ws_lf0, fn_ws_lf0, num_ws_mcp, fn_ws_mcp,
                      rate_interp, num_interp);

   /* generate speech */
   HTS_Model_load_from_labfp(&m, labfp);
   HTS_Engine_process(&engine, &m,
                      wavfp, rawfp, lf0fp, mcpfp, durfp, tracefp, NULL, NULL);
   HTS_Engine_refresh(&engine, &m);

   /* free memory */
   HTS_Engine_clear(&engine);
   free(rate_interp);
   free(fn_ws_mcp);
   free(fn_ws_lf0);
   free(fn_ms_mcp);
   free(fn_ms_lf0);
   free(fn_ms_dur);
   free(fn_ts_mcp);
   free(fn_ts_lf0);
   free(fn_ts_dur);

   /* close files */
   if (mcpfp != NULL)
      fclose(mcpfp);
   if (lf0fp != NULL)
      fclose(lf0fp);
   if (wavfp != NULL)
      fclose(wavfp);
   if (rawfp != NULL)
      fclose(rawfp);
   if (labfp != NULL)
      fclose(labfp);
   if (durfp != NULL)
      fclose(durfp);

   return 0;
}
