/*  ---------------------------------------------------------------  */
/*           The HMM-Based Speech Synthesis System (HTS)             */
/*                       HTS Working Group                           */
/*                                                                   */
/*                  Department of Computer Science                   */
/*                  Nagoya Institute of Technology                   */
/*                               and                                 */
/*   Interdisciplinary Graduate School of Science and Engineering    */
/*                  Tokyo Institute of Technology                    */
/*                                                                   */
/*                     Copyright (c) 2001-2007                       */
/*                       All Rights Reserved.                        */
/*                                                                   */
/*  Permission is hereby granted, free of charge, to use and         */
/*  distribute this software and its documentation without           */
/*  restriction, including without limitation the rights to use,     */
/*  copy, modify, merge, publish, distribute, sublicense, and/or     */
/*  sell copies of this work, and to permit persons to whom this     */
/*  work is furnished to do so, subject to the following conditions: */
/*                                                                   */
/*    1. The source code must retain the above copyright notice,     */
/*       this list of conditions and the following disclaimer.       */
/*                                                                   */
/*    2. Any modifications to the source code must be clearly        */
/*       marked as such.                                             */
/*                                                                   */
/*    3. Redistributions in binary form must reproduce the above     */
/*       copyright notice, this list of conditions and the           */
/*       following disclaimer in the documentation and/or other      */
/*       materials provided with the distribution.  Otherwise, one   */
/*       must contact the HTS working group.                         */
/*                                                                   */
/*  NAGOYA INSTITUTE OF TECHNOLOGY, TOKYO INSTITUTE OF TECHNOLOGY,   */
/*  HTS WORKING GROUP, AND THE CONTRIBUTORS TO THIS WORK DISCLAIM    */
/*  ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL       */
/*  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT   */
/*  SHALL NAGOYA INSTITUTE OF TECHNOLOGY, TOKYO INSTITUTE OF         */
/*  TECHNOLOGY, HTS WORKING GROUP, NOR THE CONTRIBUTORS BE LIABLE    */
/*  FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY        */
/*  DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,  */
/*  WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTUOUS   */
/*  ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR          */
/*  PERFORMANCE OF THIS SOFTWARE.                                    */
/*                                                                   */
/*  ---------------------------------------------------------------  */

/* $Id: HTS_audio.c,v 1.2 2007/09/28 05:40:09 zen Exp $ */

#include "HTS_hidden.h"

/* for windows & windows mobile */
#if defined(AUDIO_PLAY_WIN32) || defined(AUDIO_PLAY_WINCE)

#define AUDIO_WAIT_BUFF_MS 10   /* wait time (ms) */
#define AUDIO_CHANNEL 1         /* monoral */

/* HTS_AudioSet_callback_function: callback function from audio device */
static void CALLBACK HTS_AudioSet_callback_function(HWAVEOUT hwaveout, UINT msg,
                                                    DWORD user_data,
                                                    DWORD param1, DWORD param2)
{
   WAVEHDR *wavehdr = (WAVEHDR *) param1;
   HTS_AudioSet *as = (HTS_AudioSet *) user_data;

   if (msg == MM_WOM_DONE && wavehdr && (wavehdr->dwFlags & WHDR_DONE)) {
      if (as->now_buff_1 == TRUE && wavehdr == &(as->buff_1)) {
         as->now_buff_1 = FALSE;
      } else if (as->now_buff_2 == TRUE && wavehdr == &(as->buff_2)) {
         as->now_buff_2 = FALSE;
      }
   }
}

/* HTS_AudioSet_write_buffer: send buffer to audio device */
static void HTS_AudioSet_write_buffer(HTS_AudioSet * as)
{
   MMRESULT error;

   if (as->which_buff == 1) {
      while (as->now_buff_1 == TRUE)
         Sleep(AUDIO_WAIT_BUFF_MS);
      as->now_buff_1 = TRUE;
      as->which_buff = 2;
      memcpy(as->buff_1.lpData, as->buff, as->buff_size * sizeof(short));
      as->buff_1.dwBufferLength = as->buff_size * sizeof(short);
      error = waveOutWrite(as->hwaveout, &(as->buff_1), sizeof(WAVEHDR));
   } else {
      while (as->now_buff_2 == TRUE)
         Sleep(AUDIO_WAIT_BUFF_MS);
      as->now_buff_2 = TRUE;
      as->which_buff = 1;
      memcpy(as->buff_2.lpData, as->buff, as->buff_size * sizeof(short));
      as->buff_2.dwBufferLength = as->buff_size * sizeof(short);
      error = waveOutWrite(as->hwaveout, &(as->buff_2), sizeof(WAVEHDR));
   }

   if (error != MMSYSERR_NOERROR)
      HTS_error(0, "hts_engine: can't send buffer.\n");
}

/* HTS_AudioSet_open: open audio device */
void HTS_AudioSet_open(HTS_AudioSet * as, int sampling_rate, int max_buff_size)
{
   MMRESULT error;

   /* queue */
   as->which_buff = 1;
   as->now_buff_1 = FALSE;
   as->now_buff_2 = FALSE;
   as->max_buff_size = max_buff_size;
   as->buff = (short *) HTS_calloc(max_buff_size, sizeof(short));

   /* format */
   as->waveformatex.wFormatTag = WAVE_FORMAT_PCM;
   as->waveformatex.nChannels = AUDIO_CHANNEL;
   as->waveformatex.nSamplesPerSec = sampling_rate;
   as->waveformatex.wBitsPerSample = sizeof(short) * 8;
   as->waveformatex.nBlockAlign = AUDIO_CHANNEL
       * as->waveformatex.wBitsPerSample / 8;
   as->waveformatex.nAvgBytesPerSec = sampling_rate
       * as->waveformatex.nBlockAlign;
   /* open */
   error =
       waveOutOpen(&as->hwaveout, WAVE_MAPPER, &as->waveformatex,
                   (DWORD) HTS_AudioSet_callback_function, (DWORD) as,
                   CALLBACK_FUNCTION);
   if (error != MMSYSERR_NOERROR)
      HTS_error(0, "hts_engine: can't open audio device.\n");

   /* prepare */
   as->buff_1.lpData = (LPSTR) HTS_calloc(max_buff_size, sizeof(short));
   as->buff_1.dwBufferLength = max_buff_size * sizeof(short);
   as->buff_1.dwFlags = WHDR_BEGINLOOP | WHDR_ENDLOOP;
   as->buff_1.dwLoops = 1;
   as->buff_1.lpNext = 0;
   as->buff_1.reserved = 0;
   error = waveOutPrepareHeader(as->hwaveout, &(as->buff_1), sizeof(WAVEHDR));
   if (error != MMSYSERR_NOERROR)
      HTS_error(0, "hts_engine: can't prepare audio buffer.\n");
   as->buff_2.lpData = (LPSTR) HTS_calloc(max_buff_size, sizeof(short));
   as->buff_2.dwBufferLength = max_buff_size * sizeof(short);
   as->buff_2.dwFlags = WHDR_BEGINLOOP | WHDR_ENDLOOP;
   as->buff_2.dwLoops = 1;
   as->buff_2.lpNext = 0;
   as->buff_2.reserved = 0;
   error = waveOutPrepareHeader(as->hwaveout, &(as->buff_2), sizeof(WAVEHDR));
   if (error != MMSYSERR_NOERROR)
      HTS_error(0, "hts_engine: can't prepare audio buffer.\n");
}

/* HTS_AudioSet_write: set data to buffer */
void HTS_AudioSet_write(HTS_AudioSet * as, short data)
{
   as->buff[as->buff_size] = data;
   as->buff_size++;
   if (as->buff_size == as->max_buff_size) {
      HTS_AudioSet_write_buffer(as);
      as->buff_size = 0;
   }
}

/* HTS_AudioSet_close: close audio device */
void HTS_AudioSet_close(HTS_AudioSet * as)
{
   MMRESULT error;

   if (as->buff_size != 0)
      HTS_AudioSet_write_buffer(as);
   while (as->now_buff_1 == TRUE)
      Sleep(AUDIO_WAIT_BUFF_MS);
   while (as->now_buff_2 == TRUE)
      Sleep(AUDIO_WAIT_BUFF_MS);
   /* stop audio */
   error = waveOutReset(as->hwaveout);
   if (error != MMSYSERR_NOERROR)
      HTS_error(0, "hts_engine: can't stop audio device.\n");
   /* unprepare */
   error = waveOutUnprepareHeader(as->hwaveout, &(as->buff_1), sizeof(WAVEHDR));
   if (error != MMSYSERR_NOERROR)
      HTS_error(0, "hts_engine: can't unprepare audio buffer.\n");
   error = waveOutUnprepareHeader(as->hwaveout, &(as->buff_2), sizeof(WAVEHDR));
   if (error != MMSYSERR_NOERROR)
      HTS_error(0, "hts_engine: can't unprepare audio buffer.\n");
   /* close */
   error = waveOutClose(as->hwaveout);
   if (error != MMSYSERR_NOERROR)
      HTS_error(0, "hts_engine: can't close audio device.\n");
   HTS_free(as->buff_1.lpData);
   HTS_free(as->buff_2.lpData);
   HTS_free(as->buff);
}
#endif                          /* AUDIO_PLAY_WIN32 || AUDIO_PLAY_WINCE */

/* for ALSA on Linux */
#ifdef AUDIO_PLAY_ALSA
void HTS_AudioSet_open(HTS_AudioSet * as, int sampling_rate, int max_buff_size)
{
}
void HTS_AudioSet_write(HTS_AudioSet * as, short data)
{
}
void HTS_AudioSet_close(HTS_AudioSet * as)
{
}
#endif                          /* AUDIO_PLAY_ALSA */

/* for others */
#ifdef AUDIO_PLAY_NONE
void HTS_AudioSet_open(HTS_AudioSet * as, int sampling_rate, int max_buff_size)
{
}
void HTS_AudioSet_write(HTS_AudioSet * as, short data)
{
}
void HTS_AudioSet_close(HTS_AudioSet * as)
{
}
#endif                          /* AUDIO_PLAY_NONE */
