/*  ---------------------------------------------------------------  */
/*           The HMM-Based Speech Synthesis System (HTS)             */
/*                       HTS Working Group                           */
/*                                                                   */
/*                  Department of Computer Science                   */
/*                  Nagoya Institute of Technology                   */
/*                               and                                 */
/*   Interdisciplinary Graduate School of Science and Engineering    */
/*                  Tokyo Institute of Technology                    */
/*                                                                   */
/*                     Copyright (c) 2001-2007                       */
/*                       All Rights Reserved.                        */
/*                                                                   */
/*  Permission is hereby granted, free of charge, to use and         */
/*  distribute this software and its documentation without           */
/*  restriction, including without limitation the rights to use,     */
/*  copy, modify, merge, publish, distribute, sublicense, and/or     */
/*  sell copies of this work, and to permit persons to whom this     */
/*  work is furnished to do so, subject to the following conditions: */
/*                                                                   */
/*    1. The source code must retain the above copyright notice,     */
/*       this list of conditions and the following disclaimer.       */
/*                                                                   */
/*    2. Any modifications to the source code must be clearly        */
/*       marked as such.                                             */
/*                                                                   */
/*    3. Redistributions in binary form must reproduce the above     */
/*       copyright notice, this list of conditions and the           */
/*       following disclaimer in the documentation and/or other      */
/*       materials provided with the distribution.  Otherwise, one   */
/*       must contact the HTS working group.                         */
/*                                                                   */
/*  NAGOYA INSTITUTE OF TECHNOLOGY, TOKYO INSTITUTE OF TECHNOLOGY,   */
/*  HTS WORKING GROUP, AND THE CONTRIBUTORS TO THIS WORK DISCLAIM    */
/*  ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL       */
/*  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT   */
/*  SHALL NAGOYA INSTITUTE OF TECHNOLOGY, TOKYO INSTITUTE OF         */
/*  TECHNOLOGY, HTS WORKING GROUP, NOR THE CONTRIBUTORS BE LIABLE    */
/*  FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY        */
/*  DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,  */
/*  WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTUOUS   */
/*  ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR          */
/*  PERFORMANCE OF THIS SOFTWARE.                                    */
/*                                                                   */
/*  ---------------------------------------------------------------  */

/* $Id: HTS_engine_lib.c,v 1.2 2007/09/28 05:40:09 zen Exp $ */

#include <stdlib.h>
#include <ctype.h>
#include <math.h>

#include "HTS_hidden.h"

/* HTS_Engine_load_fn: setup by filenames */
void HTS_Engine_load_fn(HTS_Engine * engine,
                        char **fn_ms_lf0, char **fn_ms_mcp,
                        char **fn_ms_dur, char **fn_ts_lf0,
                        char **fn_ts_mcp, char **fn_ts_dur,
                        int num_ws_lf0, char **fn_ws_lf0,
                        int num_ws_mcp, char **fn_ws_mcp,
                        double *rate_interp, int num_interp)
{
   int i;
   FILE **fp_ms_lf0, **fp_ms_mcp, **fp_ms_dur;
   FILE **fp_ts_lf0, **fp_ts_mcp, **fp_ts_dur;
   FILE **fp_ws_lf0, **fp_ws_mcp;

   /* initialize */
   fp_ms_lf0 = (FILE **) HTS_calloc(num_interp, sizeof(FILE *));
   fp_ms_mcp = (FILE **) HTS_calloc(num_interp, sizeof(FILE *));
   fp_ms_dur = (FILE **) HTS_calloc(num_interp, sizeof(FILE *));
   fp_ts_lf0 = (FILE **) HTS_calloc(num_interp, sizeof(FILE *));
   fp_ts_mcp = (FILE **) HTS_calloc(num_interp, sizeof(FILE *));
   fp_ts_dur = (FILE **) HTS_calloc(num_interp, sizeof(FILE *));
   fp_ws_lf0 = (FILE **) HTS_calloc(num_ws_lf0, sizeof(FILE *));
   fp_ws_mcp = (FILE **) HTS_calloc(num_ws_mcp, sizeof(FILE *));
   /* open file pointers */
   for (i = 0; i < num_interp; i++) {
      fp_ms_lf0[i] = HTS_get_fp(fn_ms_lf0[i], "rb");
      fp_ms_mcp[i] = HTS_get_fp(fn_ms_mcp[i], "rb");
      fp_ms_dur[i] = HTS_get_fp(fn_ms_dur[i], "rb");
      fp_ts_lf0[i] = HTS_get_fp(fn_ts_lf0[i], "r");
      fp_ts_mcp[i] = HTS_get_fp(fn_ts_mcp[i], "r");
      fp_ts_dur[i] = HTS_get_fp(fn_ts_dur[i], "r");
   }
   for (i = 0; i < num_ws_lf0; i++)
      fp_ws_lf0[i] = HTS_get_fp(fn_ws_lf0[i], "r");
   for (i = 0; i < num_ws_mcp; i++)
      fp_ws_mcp[i] = HTS_get_fp(fn_ws_mcp[i], "r");

   HTS_Engine_load_fp(engine,
                      fp_ms_lf0, fp_ms_mcp, fp_ms_dur, fp_ts_lf0,
                      fp_ts_mcp, fp_ts_dur, num_ws_lf0, fp_ws_lf0,
                      num_ws_mcp, fp_ws_mcp, rate_interp, num_interp);

   /* close file pointers */
   for (i = 0; i < num_interp; i++) {
      fclose(fp_ms_lf0[i]);
      fclose(fp_ms_mcp[i]);
      fclose(fp_ms_dur[i]);
      fclose(fp_ts_lf0[i]);
      fclose(fp_ts_mcp[i]);
      fclose(fp_ts_dur[i]);
   }
   for (i = 0; i < num_ws_lf0; i++)
      fclose(fp_ws_lf0[i]);
   for (i = 0; i < num_ws_mcp; i++)
      fclose(fp_ws_mcp[i]);
   /* free */
   HTS_free(fp_ms_lf0);
   HTS_free(fp_ms_mcp);
   HTS_free(fp_ms_dur);
   HTS_free(fp_ts_lf0);
   HTS_free(fp_ts_mcp);
   HTS_free(fp_ts_dur);
   HTS_free(fp_ws_lf0);
   HTS_free(fp_ws_mcp);
}

/* HTS_Engine_load_fp: setup by file pointers */
void HTS_Engine_load_fp(HTS_Engine * engine,
                        FILE ** fp_ms_lf0, FILE ** fp_ms_mcp,
                        FILE ** fp_ms_dur, FILE ** fp_ts_lf0,
                        FILE ** fp_ts_mcp, FILE ** fp_ts_dur,
                        int num_ws_lf0, FILE ** fp_ws_lf0,
                        int num_ws_mcp, FILE ** fp_ws_mcp,
                        double *rate_interp, int num_interp)
{
   int i;
   double sum = 0.0;
   if (num_interp < 1)
      HTS_error(1, "hts_engine: num_interp is too small\n");

   /* initialize an array of HTS_ModelSet,TreeSet for interpolation */
   engine->num_interp = num_interp;
   engine->ms = (HTS_ModelSet *) HTS_calloc(num_interp, sizeof(HTS_ModelSet));
   engine->ts = (HTS_TreeSet *) HTS_calloc(num_interp, sizeof(HTS_TreeSet));
   if (rate_interp == NULL) {
      for (i = 0; i < num_interp; i++) {
         engine->ms[i].weight_interp = 1 / num_interp;
      }
   } else {
      for (i = 0; i < num_interp; i++) {
         sum += rate_interp[i];
      }
      for (i = 0; i < num_interp; i++) {
         engine->ms[i].weight_interp = rate_interp[i] / sum;
      }
   }
   for (i = 0; i < num_interp; i++) {
      HTS_HTS_ModelSet_load(&engine->ms[i], fp_ms_lf0[i], fp_ms_mcp[i],
                            fp_ms_dur[i]);
      HTS_TreeSet_load(&engine->ts[i], fp_ts_lf0[i], LF0);
      HTS_TreeSet_load(&engine->ts[i], fp_ts_mcp[i], MCP);
      HTS_TreeSet_load(&engine->ts[i], fp_ts_dur[i], DUR);
   }
   /* check */
   if (num_ws_lf0 != engine->ms[0].lf0stream) {
      HTS_error(1,
                "hts_engine: #window for log f0 is not matched to acoustic model.\n");
   }
   if (engine->ms[0].mcpvsize % num_ws_mcp != 0) {
      HTS_error(1,
                "hts_engine: #window for mel-cepstrum is not matched to acoustic model.\n");
   }
   /* initialize windows */
   HTS_PStream_load(&engine->lf0pst, &engine->mcppst,
                    fp_ws_lf0, fp_ws_mcp, num_ws_lf0, num_ws_mcp,
                    engine->ms[0].mcpvsize);
}

/* HTS_Engine_initialize: initialize hts_engine */
void HTS_Engine_initialize(HTS_Engine * engine)
{
   /* default value for control parameter */
   engine->gp.rate = 16000;
   engine->gp.fperiod = 80;
   engine->gp.rho = 0.0;
   engine->gp.alpha = 0.42;
   engine->gp.f0_std = 1.0;
   engine->gp.f0_mean = 0.0;
   engine->gp.uv = 0.5;
   engine->gp.beta = 0.0;
   engine->gp.length = 0.0;
   engine->gp.algnst = FALSE;
   engine->gp.algnph = FALSE;

   engine->gp.buff_size = 0;

   engine->gp.nsample = 0;
   engine->gp.totalframe = 0;

   engine->gp.stored_raw_data = FALSE;
   engine->gp.raw_data = NULL;

   engine->ms = NULL;
   engine->ts = NULL;
   engine->num_interp = 0;
}

/* HTS_Engine_set_sampling_rate: set sampling rate */
void HTS_Engine_set_sampling_rate(HTS_Engine * engine, int i)
{
   if (i > 0 && i <= 48000)
      engine->gp.rate = i;
}

/* HTS_Engine_set_fperiod: set fperiod */
void HTS_Engine_set_fperiod(HTS_Engine * engine, int i)
{
   if (i > 0 && i <= 2000)
      engine->gp.fperiod = i;
}

/* HTS_Engine_set_alpha: set alpha */
void HTS_Engine_set_alpha(HTS_Engine * engine, double f)
{
   if (f <= 1.0 && f >= 0.0)
      engine->gp.alpha = f;
}

/* HTS_Engine_set_beta: set beta */
void HTS_Engine_set_beta(HTS_Engine * engine, double f)
{
   if (f <= 0.8 && f >= -0.8)
      engine->gp.beta = f;
}

/* HTS_Engine_set_rho: set rho */
void HTS_Engine_set_rho(HTS_Engine * engine, double f)
{
   if (f <= 1.0 && f >= -1.0)
      engine->gp.rho = f;
}

/* HTS_Engine_set_f0_std: set f0_std */
void HTS_Engine_set_f0_std(HTS_Engine * engine, double f)
{
   if (f <= 5.0 && f >= 0.0)
      engine->gp.f0_std = f;
}

/* HTS_Engine_set_f0_mean: set f0_mean */
void HTS_Engine_set_f0_mean(HTS_Engine * engine, double f)
{
   if (f <= 100.0 && f >= 0.0)
      engine->gp.f0_mean = f;
}

/* HTS_Engine_set_uv: set uv */
void HTS_Engine_set_uv(HTS_Engine * engine, double f)
{
   if (f <= 1.0 && f >= 0.0)
      engine->gp.uv = f;
}

/* HTS_Engine_set_length: set LENGTH */
void HTS_Engine_set_length(HTS_Engine * engine, double f)
{
   if (f <= 30.0 && f >= 0.0)
      engine->gp.length = f;
}

/* HTS_Engine_set_algnst: set algnst */
void HTS_Engine_set_algnst(HTS_Engine * engine, HTS_Boolean i)
{
   engine->gp.algnst = i;
}

/* HTS_Engine_set_algnph: set algnph */
void HTS_Engine_set_algnph(HTS_Engine * engine, HTS_Boolean i)
{
   engine->gp.algnph = i;
}

/* HTS_Engine_set_buff_size: set audio buffer size */
void HTS_Engine_set_buff_size(HTS_Engine * engine, int i)
{
   if (0 < i && i <= 48000)
      engine->gp.buff_size = i;
}

/* HTS_Engine_get_sampling_rate: get sampling rate */
int HTS_Engine_get_sampling_rate(HTS_Engine * engine)
{
   return engine->gp.rate;
}

/* HTS_Engine_get_total_dur: get totaldur */
int HTS_Engine_get_total_dur(HTS_Engine * engine)
{
   return engine->gp.totaldur;
}

/* HTS_Engine_get_total_frame: get totalframe */
int HTS_Engine_get_total_frame(HTS_Engine * engine)
{
   return engine->gp.totalframe;
}

/* HTS_GetNSample: get nsample */
int HTS_Engine_get_nsample(HTS_Engine * engine)
{
   return engine->gp.nsample;
}

/* HTS_Engine_clear: free memory */
void HTS_Engine_clear(HTS_Engine * engine)
{
   int i;
   /* free memory */
   for (i = 0; i < engine->num_interp; i++) {
      HTS_HTS_ModelSet_clear(&engine->ms[i]);
      HTS_TreeSet_clear(&engine->ts[i], LF0);
      HTS_TreeSet_clear(&engine->ts[i], MCP);
      HTS_TreeSet_clear(&engine->ts[i], DUR);
   }
   HTS_PStream_clear(&engine->lf0pst, &engine->mcppst);

   HTS_free(engine->ms);
   HTS_free(engine->ts);
   engine->ms = NULL;
   engine->ts = NULL;
   engine->num_interp = 0;
}

/* HTS_Engine_refresh: free model */
void HTS_Engine_refresh(HTS_Engine * engine, HTS_Model * mhead)
{
   HTS_Model *m, *next;
   int state;
   HTS_ModelSet *ms = engine->ms;

   /* free memory */
   HTS_PStream_refresh(&engine->lf0pst, &engine->mcppst);
   for (m = mhead; m; m = next) {
      for (state = ms[0].nstate + 1; state >= 2; state--) {
         HTS_free(m->mcpmean[state]);
         HTS_free(m->mcpvariance[state]);
      }

      next = m->next;
      for (state = ms[0].nstate + 1; state >= 2; state--) {
         if (m->lf0variance != NULL)
            HTS_free(++m->lf0variance[state]);
         if (m->lf0mean != NULL)
            HTS_free(++m->lf0mean[state]);
      }

      m->dur += 2;
      m->lf0pdf += 2;
      m->lf0mean += 2;
      m->lf0variance += 2;
      m->voiced += 2;
      m->mcppdf += 2;
      m->mcpmean += 2;
      m->mcpvariance += 2;

      HTS_free(m->lf0variance);
      HTS_free(m->lf0mean);
      HTS_free(m->lf0pdf);
      HTS_free(m->mcpvariance);
      HTS_free(m->mcpmean);
      HTS_free(m->mcppdf);
      HTS_free(m->voiced);
      HTS_free(m->dur);
      HTS_free(m->name);
      HTS_free(m->lab);
      if (m != mhead)
         HTS_free(m);
   }

   engine->gp.nsample = 0;
   engine->gp.totalframe = 0;
   if (engine->gp.raw_data != NULL)
      HTS_free(engine->gp.raw_data);
   engine->gp.raw_data = NULL;
}

/* OutLabel: output label with frame number or time */
static void HTS_OutLabel(FILE * durfp, HTS_UttModel * um, HTS_globalP * gp)
{
   HTS_Model *m;
   int i = 0, j;

   for (m = um->mhead;; m = m->next) {
      j = m->totaldur;
      /* in HTK & HTS format */
      fprintf(durfp, "%d %d %s\n",
              (int) (i * gp->fperiod * 1e+7 / gp->rate),
              (int) ((i + j) * gp->fperiod * 1e+7 / gp->rate), m->name);
      i += j;
      if (m == um->mtail)
         break;
   }
}

/* OutInfo: output trace information */
static void HTS_OutInfo(FILE * tracefp, HTS_UttModel * um, int nstate,
                        HTS_globalP * gp)
{
   HTS_Model *m;
   int i, j, t;

   /* output settings */
   fprintf(tracefp,
           "sampring frequency                     -> %d(Hz)\n", gp->rate);
   fprintf(tracefp,
           "frame period                           -> %d(point) %.2f(msec)\n",
           gp->fperiod, 1e+3 * gp->fperiod / gp->rate);
   fprintf(tracefp, "use state alignment for duration       -> %d\n",
           gp->algnst);
   fprintf(tracefp, "use phoneme alignment for duration     -> %d\n",
           gp->algnph);
   fprintf(tracefp, "all-pass constant                      -> %f\n",
           gp->alpha);
   fprintf(tracefp, "postfiltering coefficient              -> %f\n", gp->beta);
   fprintf(tracefp, "control duration parameter             -> %f\n", gp->rho);
   fprintf(tracefp, "multilply f0                           -> %f\n",
           gp->f0_std);
   fprintf(tracefp, "add f0                                 -> %f\n",
           gp->f0_mean);
   fprintf(tracefp, "voiced/unvoiced threshold              -> %f\n", gp->uv);
   fprintf(tracefp,
           "specified utterance length             -> %f(sec.)\n\n",
           gp->length);

   /* output sentence HMM and generated utterance information */
   fprintf(tracefp, "number of HMMs        -> %d\n", um->nModel);
   fprintf(tracefp, "number of HMM states  -> %d\n", um->nState);
   fprintf(tracefp,
           "length of this speech -> %1.3lf sec. (%d frames)\n",
           (double) (um->totalframe * gp->fperiod) / gp->rate, um->totalframe);
   fprintf(tracefp, "\n");

   /* output each state information */
   for (m = um->mhead, j = 1, t = 0;; m = m->next, j++) {
      fprintf(tracefp, "%d: %s \n", j, m->name);        /* model name */
      fprintf(tracefp, "            duration -> %-3d\n", m->durpdf);    /* model duration */
      for (i = 2; i < nstate + 2; i++) {
         /* leaf number */
         fprintf(tracefp,
                 " %2d-state : spectrum -> %-3d   f0 -> %-3d   ", i,
                 m->mcppdf[i], m->lf0pdf[i]);
         /* state duration (time) */
         fprintf(tracefp, " %1.3lf--%1.3lf(sec)",
                 (double) (t * gp->fperiod) / gp->rate,
                 (double) ((t + m->dur[i]) * gp->fperiod) / gp->rate);
         /* state duration (#frame) */
         fprintf(tracefp, " %3d(frame)", m->dur[i]);
         /* voiced/unvoiced */
         fprintf(tracefp, "   %s\n", ((m->voiced[i]) ? "voiced" : "unvoiced"));
         t += m->dur[i];
      }
      if (m == um->mtail)
         break;
   }
}

static void HTS_add_riff_header(FILE * wavfp, int sampling_rate,
                                int number_of_point, short byte_per_point)
{
   char data_01_04[] = { 'R', 'I', 'F', 'F' };
   int data_05_08 = number_of_point * byte_per_point + 36;
   char data_09_12[] = { 'W', 'A', 'V', 'E' };
   char data_13_16[] = { 'f', 'm', 't', ' ' };
   int data_17_20 = 16;
   short data_21_22 = 1;        /* PCM */
   short data_23_24 = 1;        /* monoral */
   int data_25_28 = sampling_rate;
   int data_29_32 = sampling_rate * byte_per_point;
   short data_33_34 = byte_per_point;
   short data_35_36 = (short) byte_per_point * 8;
   char data_37_40[] = { 'd', 'a', 't', 'a' };
   int data_41_44 = number_of_point * byte_per_point;
   HTS_fwrite_little_endian(data_01_04, sizeof(char), 4, wavfp);
   HTS_fwrite_little_endian(&data_05_08, sizeof(int), 1, wavfp);
   HTS_fwrite_little_endian(data_09_12, sizeof(char), 4, wavfp);
   HTS_fwrite_little_endian(data_13_16, sizeof(char), 4, wavfp);
   HTS_fwrite_little_endian(&data_17_20, sizeof(int), 1, wavfp);
   HTS_fwrite_little_endian(&data_21_22, sizeof(short), 1, wavfp);
   HTS_fwrite_little_endian(&data_23_24, sizeof(short), 1, wavfp);
   HTS_fwrite_little_endian(&data_25_28, sizeof(int), 1, wavfp);
   HTS_fwrite_little_endian(&data_29_32, sizeof(int), 1, wavfp);
   HTS_fwrite_little_endian(&data_33_34, sizeof(short), 1, wavfp);
   HTS_fwrite_little_endian(&data_35_36, sizeof(short), 1, wavfp);
   HTS_fwrite_little_endian(data_37_40, sizeof(char), 4, wavfp);
   HTS_fwrite_little_endian(&data_41_44, sizeof(int), 1, wavfp);
}

/* HTS_Engine_process: parse label, determine state duration,
   generate sequence of speech parameter vector, and synthesize waveform */
void HTS_Engine_process(HTS_Engine * engine, HTS_Model * mhead,
                        FILE * wavfp, FILE * rawfp, FILE * lf0fp,
                        FILE * mcpfp, FILE * durfp, FILE * tracefp,
                        double *fr_f0, double *fr_power)
{
   char buf[HTS_MAXBUFLEN];
   int state;
   int start, end;
   int rate, nf;                /* nf : number of frames */
   double f, mean, var, diffdur;
   HTS_Boolean hastime;
   HTS_Model *m;
   HTS_UttModel um;
   char *lab;
   HTS_PStream *lf0pst = &engine->lf0pst;
   HTS_PStream *mcppst = &engine->mcppst;
   HTS_globalP *gp = &engine->gp;
   HTS_ModelSet *ms = engine->ms;
   HTS_TreeSet *ts = engine->ts;
   int num_interp = engine->num_interp;

   int dp = 0;                  /* data pointer */
   int totaldur;
   double *durmean, *durvar;

   /* check */
   if (engine->gp.algnst && engine->gp.algnph)
      HTS_error(1, "hts_engine: algnst and algnph are exclusive.\n");
   if (engine->gp.length > 0.0 && engine->gp.rho != 0.0)
      HTS_error(1, "hts_engine: length and rho are exclusive.\n");

   rate = gp->fperiod * 10000000 / gp->rate;

   mean = var = diffdur = 0.0;

   um.mtail = um.mhead = mhead; /* utterance model = final speaker */
   um.totalframe = um.nState = um.nModel = 0;
   start = end = 0;

   /* memory allocate for durmean & durvar (dur parameter for final speaker)  k-saino modified */
   durmean = (double *) HTS_calloc(ms[0].nstate, sizeof(double));
   durmean -= 2;
   durvar = (double *) HTS_calloc(ms[0].nstate, sizeof(double));
   durvar -= 2;

   /* parse label file */
   for (m = mhead; m; m = m->next) {
      um.mtail = m;
      lab = m->lab;
      dp = 0;
      um.nModel++;
      um.nState += ms[0].nstate;
      HTS_get_token_from_str(lab, &dp, buf);

      if (isdigit(buf[0]))
         hastime = TRUE;        /* label contain segmentation information */
      else
         hastime = FALSE;

      if (hastime) {
         if (gp->algnst) {      /* state-level segmentation */
            start = atoi(buf);  /* start time */
            HTS_get_token_from_str(lab, &dp, buf);
            end = atoi(buf);    /* end time */
            HTS_get_token_from_str(lab, &dp, buf);
            HTS_get_token_from_str(lab, &dp, buf);
         } else if (gp->algnph) {       /* phoneme-level segmentation */
            start = atoi(buf);  /* start time */
            HTS_get_token_from_str(lab, &dp, buf);
            end = atoi(buf);    /* end time */
            HTS_get_token_from_str(lab, &dp, buf);
         } else {               /* not use segmentation */
            do {
               HTS_get_token_from_str(lab, &dp, buf);
            } while (isdigit(buf[0]));
         }
      }

      /* allocate memory for current model */
      m->name = HTS_strdup(buf);        /* model name          */
      m->dur = (int *) HTS_calloc(ms[0].nstate + 2, sizeof(int));       /* duration            */
      m->voiced = (HTS_Boolean *) HTS_calloc(ms[0].nstate + 2, sizeof(HTS_Boolean));    /* voiced/unvoiced     */
      m->lf0pdf = (int *) HTS_calloc(ms[0].nstate + 2, sizeof(int));    /* f0 (leaf number)    */
      m->lf0mean = (double **) HTS_calloc(ms[0].nstate + 2, sizeof(double *));  /* f0 (mean)           */
      m->lf0variance = (double **) HTS_calloc(ms[0].nstate + 2, sizeof(double *));      /* f0 (variance)       */
      m->mcppdf = (int *) HTS_calloc(ms[0].nstate + 2, sizeof(int));    /* spectrum (leaf num) */
      m->mcpmean = (double **) HTS_calloc(ms[0].nstate + 2, sizeof(double *));  /* spectrum (mean)     */
      m->mcpvariance = (double **) HTS_calloc(ms[0].nstate + 2, sizeof(double *));      /* spectrum (variance) */

      m->dur -= 2;
      m->lf0pdf -= 2;
      m->lf0mean -= 2;
      m->lf0variance -= 2;
      m->voiced -= 2;
      m->mcppdf -= 2;
      m->mcpmean -= 2;
      m->mcpvariance -= 2;

      for (state = 2; state <= ms[0].nstate + 1; state++) {
         m->lf0mean[state] =
             (double *) HTS_calloc(ms[0].lf0stream, sizeof(double));
         m->lf0variance[state] =
             (double *) HTS_calloc(ms[0].lf0stream, sizeof(double));
         m->lf0mean[state]--;
         m->lf0variance[state]--;
      }

      /* determine state-level duration */
      if (!hastime && m->bool_dur) {    /* use duration set by user (by nkazu) */
         double diffdur, dur, tdrest, old_dur;
         totaldur = m->lab_dur;
         HTS_Model_estimate_dur(m, ms, ts, num_interp, durmean, durvar);        /* specify dur{mean,var} */
         HTS_Model_determine_dur(m, &ms[0], durmean, durvar, gp->rho, &diffdur);

         nf = 0;
         for (state = 2; state <= ms[0].nstate + 1; state++)
            nf += m->dur[state];
         f = (double) totaldur / nf;
         m->totaldur = 0;
         diffdur = 0.0;
         tdrest = totaldur;
         for (state = 2; state <= ms[0].nstate + 1; state++) {
            dur = f * m->dur[state];
            old_dur = dur;
            dur += diffdur * dur / tdrest + 0.5;
            nf = (int) dur;
            diffdur += old_dur - nf;
            tdrest -= old_dur;
            m->dur[state] = nf;
            m->totaldur += m->dur[state];
         }
         um.totalframe += m->totaldur;
      } else if (!hastime && m->bool_rate) {    /* use rate set by user (by nkazu) */
         HTS_Model_estimate_dur(m, ms, ts, num_interp, durmean, durvar);        /* specify dur{mean,var} */
         HTS_Model_determine_dur(m, &ms[0], durmean, durvar, gp->rho, &diffdur);

         m->totaldur = 0;
         for (state = 2; state <= ms[0].nstate + 1; state++) {
            nf = (int) (m->lab_rate * m->dur[state] + 0.5);
            if (nf <= 0)
               nf = 1;
            m->dur[state] = nf;
            m->totaldur += m->dur[state];
         }
         um.totalframe += m->totaldur;
      } else if (hastime && gp->algnph) {       /* use phoneme-level segmentation */
         HTS_Model_estimate_dur(m, ms, ts, num_interp, durmean, durvar);        /* specify dur{mean,var} */
         HTS_Model_determine_dur(m, &ms[0], durmean, durvar, gp->rho, &diffdur);

         nf = 0;
         for (state = 2; state <= ms[0].nstate + 1; state++)
            nf += m->dur[state];

         fprintf(stderr, ">>>nf=%d %d\n", nf, (end - start) / rate);

         f = (double) (end - start) / (rate * nf);
         m->totaldur = 0;

         for (state = 2; state <= ms[0].nstate + 1; state++) {
            nf = (int) (f * m->dur[state] + 0.5);
            if (nf <= 0)
               nf = 1;
            fprintf(stderr, "%d: %d %f %d\n", state, m->dur[state], f, nf);
            m->dur[state] = nf;
            m->totaldur += m->dur[state];
         }
         um.totalframe += m->totaldur;
      } else if (hastime && gp->algnst) {       /* use state-level segmentation */
         m->dur[2] = (end - start) / rate;
         m->totaldur = m->dur[2];
         um.totalframe += m->dur[2];

         for (state = 3; state <= ms[0].nstate + 1; state++) {
            HTS_get_token_from_str(lab, &dp, buf);
            start = atoi(buf);
            HTS_get_token_from_str(lab, &dp, buf);
            end = atoi(buf);
            HTS_get_token_from_str(lab, &dp, buf);
            m->dur[state] = (end - start) / rate;
            m->totaldur += m->dur[state];
            um.totalframe += m->dur[state];
         }
      } else {                  /* estimate state duration from state duration model (Gaussian) */
         HTS_Model_estimate_dur(m, ms, ts, num_interp, durmean, durvar);        /* specify dur{mean,var} */
         if (gp->length == 0.0) {
            HTS_Model_determine_dur(m, &ms[0], durmean, durvar,
                                    gp->rho, &diffdur);
            um.totalframe += m->totaldur;
         } else {               /* if total length of generated speech is specified */
            for (state = 2; state <= ms[0].nstate + 1; state++) {
               mean += durmean[state];
               var += durvar[state];
            }
         }
      }

      /* find pdf for lf0 & interpolate */
      HTS_Model_determine_lf0(m, ms, ts, num_interp, gp->uv);
      /* mel-cepstrum interpolation */
      HTS_Model_determine_mcp(m, ms, ts, num_interp);

   }

   /* Specified utterance length is too short */
   if (gp->length > 0.0 && gp->length * gp->rate / gp->fperiod < um.nState)
      HTS_error(1, "hts_engine: specified utterance length is too short.\n");

   /* if total length of utterance is specified, RHO (temporal factor) have to be computed */
   if (gp->length > 0.0) {
      gp->rho = (((int) gp->length * gp->rate / gp->fperiod) - mean) / var;
      /* compute state duration for each state */
      for (m = um.mhead;; m = m->next) {
         HTS_Model_estimate_dur(m, ms, ts, num_interp, durmean, durvar);        /* specify dur{mean,var} */
         HTS_Model_determine_dur(m, &ms[0], durmean, durvar, gp->rho, &diffdur);

         um.totalframe += m->totaldur;
         if (m == um.mtail)
            break;
      }
   }

   HTS_free(durmean + 2);
   HTS_free(durvar + 2);

   /* output trace information */
   if (tracefp != NULL)
      HTS_OutInfo(tracefp, &um, ms[0].nstate, gp);

   /* output segment information */
   if (durfp != NULL)
      HTS_OutLabel(durfp, &um, gp);

   gp->nsample = um.totalframe * gp->fperiod;
   gp->totalframe = um.totalframe;
   gp->totaldur = um.mhead->totaldur;

   if (wavfp != NULL)
      HTS_add_riff_header(wavfp, gp->rate, gp->nsample, sizeof(short));

   /* generate speech parameter vector sequence and synthesize speech waveform */
   HTS_pdf2speech(wavfp, rawfp, lf0fp, mcpfp, fr_f0, fr_power,
                  lf0pst, mcppst, gp, &ms[0], &um);
}

/* HTS_Model_load_from_labfn: create model list from file name */
void HTS_Model_load_from_labfn(HTS_Model * mhead, char *fn)
{
   FILE *fp = HTS_get_fp(fn, "r");
   HTS_Model_load_from_labfp(mhead, fp);
   fclose(fp);
}

/* HTS_Model_load_from_labfp: create model list from file pointer */
void HTS_Model_load_from_labfp(HTS_Model * mhead, FILE * labfp)
{
   char buf[HTS_MAXBUFLEN];
   HTS_Model *m = NULL;

   /* parse label file */
   while (HTS_get_line(labfp, buf)) {
      if (!isgraph(buf[0]))
         break;
      if (m) {
         m->next = (HTS_Model *) HTS_calloc(1, sizeof(HTS_Model));
         m = m->next;
      } else {                  /* first time */
         m = mhead;
      }
      m->lab = HTS_strdup(buf);
      m->totaldur = 0;

      /* for galatea */
      m->bool_rate = FALSE;
      m->bool_dur = FALSE;
      m->bool_f0_level = FALSE;
      m->bool_f0_range = FALSE;
      m->bool_volume = FALSE;
      m->bool_alpha = FALSE;
   }
   m->next = NULL;
}

/* HTS_Model_load_from_string: create model list from string */
void HTS_Model_load_from_string(HTS_Model * mhead, char *labdata)
{
   char buf[HTS_MAXBUFLEN];
   HTS_Model *m = NULL;

   int dp = 0;                  /* data pointer */

   /* parse label file */
   while (HTS_get_line_from_str(labdata, &dp, buf)) {
      if (!isgraph(buf[0]))
         break;
      if (m) {
         m->next = (HTS_Model *) HTS_calloc(1, sizeof(HTS_Model));
         m = m->next;
      } else {                  /* first time */
         m = mhead;
      }
      m->lab = HTS_strdup(buf);
      m->totaldur = 0;

      /* for galatea */
      m->bool_rate = FALSE;
      m->bool_dur = FALSE;
      m->bool_f0_level = FALSE;
      m->bool_f0_range = FALSE;
      m->bool_volume = FALSE;
      m->bool_alpha = FALSE;
   }
   m->next = NULL;
}

/* HTS_Model_load_from_string_list: create model list from string list */
void HTS_Model_load_from_string_list(HTS_Model * mhead, char **labdata,
                                     int size)
{
   int i;
   HTS_Model *m = NULL;

   /* parse label file */
   for (i = 0; i < size; i++) {
      if (!isgraph(labdata[i][0]))
         break;
      if (m) {
         m->next = (HTS_Model *) HTS_calloc(1, sizeof(HTS_Model));
         m = m->next;
      } else {                  /* first time */
         m = mhead;
      }
      m->lab = HTS_strdup(labdata[i]);
      m->totaldur = 0;
      /* for galatea */
      m->bool_rate = FALSE;
      m->bool_dur = FALSE;
      m->bool_f0_level = FALSE;
      m->bool_f0_range = FALSE;
      m->bool_volume = FALSE;
      m->bool_alpha = FALSE;
   }
   m->next = NULL;
}

/* HTS_Model_set_rate: set model rate */
void HTS_Model_set_rate(HTS_Model * m, int index, double f)
{
   while (index--)
      m = m->next;
   if (f < 0.2)
      f = 0.2;
   m->lab_rate = f;
   m->bool_rate = TRUE;
}

/* HTS_Model_set_dur: set model duration */
void HTS_Model_set_dur(HTS_Model * m, int index, int i)
{
   while (index--)
      m = m->next;
   if (i < 0)
      i = 0;
   m->lab_dur = i;
   m->bool_dur = TRUE;
}

/* HTS_Model_set_f0_level: set model f0 level */
void HTS_Model_set_f0_level(HTS_Model * m, int index, double f)
{
   while (index--)
      m = m->next;
   if (f < 0.1)
      f = 0.1;
   m->lab_f0_level = f;
   m->bool_f0_level = TRUE;
}

/* HTS_Model_set_f0_range: set model f0 range */
void HTS_Model_set_f0_range(HTS_Model * m, int index, double f)
{
   while (index--)
      m = m->next;
   if (f < 0.0)
      f = 0.0;
   m->lab_f0_range = f;
   m->bool_f0_range = TRUE;
}

/* HTS_Model_set_volume: set model volume */
void HTS_Model_set_volume(HTS_Model * m, int index, double f)
{
   while (index--)
      m = m->next;
   if (f < 0.01)
      f = 0.01;
   m->lab_volume = f;
   m->bool_volume = TRUE;
}

/* HTS_Model_set_alpha: set model alpha */
void HTS_Model_set_alpha(HTS_Model * m, int index, double f)
{
   while (index--)
      m = m->next;
   if (f >= 0.0) {
      m->lab_alpha = f;
      m->bool_alpha = TRUE;
   }
}

/* HTS_Engine_get_pros_len: get pros length */
int HTS_Engine_get_pros_len(HTS_Engine * engine)
{
   return engine->mcppst.T;
}

/* HTS_Engine_get_pros: get pros */
HTS_Boolean HTS_Engine_get_pros(HTS_Engine * engine,
                                int len, double *f0_data, double *power_data)
{
   int lf0frame, mcpframe;
   double f0;
   HTS_PStream *lf0pst = &engine->lf0pst;
   HTS_PStream *mcppst = &engine->mcppst;
   HTS_globalP *gp = &engine->gp;

   /* check length */
   if (mcppst->T != len) {
      return FALSE;
   }

   for (lf0frame = mcpframe = 1; mcpframe <= len; mcpframe++) {
      power_data[mcpframe] = mcppst->par[mcpframe][1];
      if (lf0pst->voiced[mcpframe])
         f0 = gp->f0_std * exp(lf0pst->par[lf0frame++][1]) + gp->f0_mean;
      else
         f0 = 0.0;
      f0_data[mcpframe] = (f0 == 0.0) ? 0.0 : log(f0);
   }
   return TRUE;
}

void HTS_Engine_set_stored_raw_data(HTS_Engine * engine, HTS_Boolean i)
{
   engine->gp.stored_raw_data = i;
}

short *HTS_Engine_get_stored_raw_data(HTS_Engine * engine)
{
   return engine->gp.raw_data;
}
